# mypy: ignore-errors
"""
=================
Results Interface
=================

This module provides a :class:`ResultsInterface <ResultsInterface>` class with
methods to register stratifications and results producers (referred to as "observations")
to a simulation.

"""

from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

import pandas as pd

from vivarium.framework.event import Event
from vivarium.framework.results.observation import (
    AddingObservation,
    ConcatenatingObservation,
    StratifiedObservation,
    UnstratifiedObservation,
)
from vivarium.manager import Interface
from vivarium.types import ScalarValue

if TYPE_CHECKING:
    from vivarium.framework.results.manager import ResultsManager


def _required_function_placeholder(*args, **kwargs) -> pd.DataFrame:
    """Placeholder function to indicate that a required function is missing."""
    return pd.DataFrame()


class ResultsInterface(Interface):
    """Builder interface for the results management system.

    The results management system allows users to delegate results production
    to the simulation framework. This process attempts to roughly mimic the
    groupby-apply logic commonly done when manipulating :mod:`pandas`
    DataFrames. The representation of state in the simulation is complex,
    however, as it includes information both in the population state table
    and dynamically generated information available from the
    :class:`value pipelines <vivarium.framework.values.Pipeline>`.
    Additionally, good encapsulation of simulation logic typically has
    results production separated from the modeling code into specialized
    `Observer` components. This often highlights the need for transformations
    of the simulation state into representations that aren't needed for
    modeling, but are required for the stratification of produced results.

    The purpose of this interface is to provide controlled access to a results
    backend by means of the builder object; it exposes methods to register both
    stratifications and results producers (referred to as "observations").

    """

    def __init__(self, manager: "ResultsManager") -> None:
        self._manager: "ResultsManager" = manager
        self._name = "results_interface"

    @property
    def name(self) -> str:
        return self._name

    ##################################
    # Stratification-related methods #
    ##################################

    # TODO: It is not reflected in the sample code here, but the “when” parameter should be added
    #  to the stratification registration calls, probably as a List. Consider this after observer implementation
    def register_stratification(
        self,
        name: str,
        categories: List[str],
        excluded_categories: Optional[List[str]] = None,
        mapper: Optional[
            Union[
                Callable[[Union[pd.Series, pd.DataFrame]], pd.Series],
                Callable[[ScalarValue], str],
            ]
        ] = None,
        is_vectorized: bool = False,
        requires_columns: List[str] = [],
        requires_values: List[str] = [],
    ) -> None:
        """Registers a stratification that can be used by stratified observations.

        Parameters
        ----------
        name
            Name of the stratification.
        categories
            Exhaustive list of all possible stratification values.
        excluded_categories
            List of possible stratification values to exclude from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        mapper
            A callable that maps the columns and value pipelines specified by the
            `requires_columns` and `requires_values` arguments to the stratification
            categories. It can either map the entire population or an individual
            simulant. A simulation will fail if the `mapper` ever produces an invalid
            value.
        is_vectorized
            True if the `mapper` function will map the entire population, and False
            if it will only map a single simulant.
        requires_columns
            A list of the state table columns that are required by the `mapper`
            to produce the stratification.
        requires_values
            A list of the value pipelines that are required by the `mapper` to
            produce the stratification.
        """
        self._manager.register_stratification(
            name,
            categories,
            excluded_categories,
            mapper,
            is_vectorized,
            requires_columns,
            requires_values,
        )

    def register_binned_stratification(
        self,
        target: str,
        binned_column: str,
        bin_edges: List[Union[int, float]] = [],
        labels: List[str] = [],
        excluded_categories: Optional[List[str]] = None,
        target_type: str = "column",
        **cut_kwargs: Dict,
    ) -> None:
        """Registers a binned stratification that can be used by stratified observations.

        Parameters
        ----------
        target
            Name of the state table column or value pipeline to be binned.
        binned_column
            Name of the (binned) stratification.
        bin_edges
            List of scalars defining the bin edges, passed to :meth: pandas.cut.
            The length must be equal to the length of `labels` plus 1.
        labels
            List of string labels for bins. The length must be equal to the length
            of `bin_edges` minus 1.
        excluded_categories
            List of possible stratification values to exclude from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        target_type
            Type specification of the `target` to be binned. "column" if it's a
            state table column or "value" if it's a value pipeline.
        **cut_kwargs
            Keyword arguments for :meth: pandas.cut.
        """
        self._manager.register_binned_stratification(
            target,
            binned_column,
            bin_edges,
            labels,
            excluded_categories,
            target_type,
            **cut_kwargs,
        )

    ###############################
    # Observation-related methods #
    ###############################

    def register_stratified_observation(
        self,
        name: str,
        pop_filter: str = "tracked==True",
        when: str = "collect_metrics",
        requires_columns: List[str] = [],
        requires_values: List[str] = [],
        results_updater: Callable[
            [pd.DataFrame, pd.DataFrame], pd.DataFrame
        ] = _required_function_placeholder,
        results_formatter: Callable[
            [str, pd.DataFrame], pd.DataFrame
        ] = lambda measure, results: results,
        additional_stratifications: List[str] = [],
        excluded_stratifications: List[str] = [],
        aggregator_sources: Optional[List[str]] = None,
        aggregator: Callable[[pd.DataFrame], Union[float, pd.Series]] = len,
        to_observe: Callable[[Event], bool] = lambda event: True,
    ) -> None:
        """Registers a stratified observation to the results system.

        Parameters
        ----------
        name
            Name of the observation. It will also be the name of the output results file
            for this particular observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        when
            Name of the lifecycle phase the observation should happen. Valid values are:
            "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
        requires_columns
            List of the state table columns that are required by either the `pop_filter` or the `aggregator`.
        requires_values
            List of the value pipelines that are required by either the `pop_filter` or the `aggregator`.
        results_updater
            Function that updates existing raw observation results with newly gathered results.
        results_formatter
            Function that formats the raw observation results.
        additional_stratifications
            List of additional :class:`Stratification <vivarium.framework.results.stratification.Stratification>`
            names by which to stratify this observation by.
        excluded_stratifications
            List of default :class:`Stratification <vivarium.framework.results.stratification.Stratification>`
            names to remove from this observation.
        aggregator_sources
            List of population view columns to be used in the `aggregator`.
        aggregator
            Function that computes the quantity for this observation.
        to_observe
            Function that determines whether to perform an observation on this Event.

        Raises
        ------
        ValueError
            If any required callable arguments are missing.
        """
        self._check_for_required_callables(name, {"results_updater": results_updater})
        self._manager.register_observation(
            observation_type=StratifiedObservation,
            is_stratified=True,
            name=name,
            pop_filter=pop_filter,
            when=when,
            requires_columns=requires_columns,
            requires_values=requires_values,
            results_updater=results_updater,
            results_formatter=results_formatter,
            additional_stratifications=additional_stratifications,
            excluded_stratifications=excluded_stratifications,
            aggregator_sources=aggregator_sources,
            aggregator=aggregator,
            to_observe=to_observe,
        )

    def register_unstratified_observation(
        self,
        name: str,
        pop_filter: str = "tracked==True",
        when: str = "collect_metrics",
        requires_columns: List[str] = [],
        requires_values: List[str] = [],
        results_gatherer: Callable[
            [pd.DataFrame], pd.DataFrame
        ] = _required_function_placeholder,
        results_updater: Callable[
            [pd.DataFrame, pd.DataFrame], pd.DataFrame
        ] = _required_function_placeholder,
        results_formatter: Callable[
            [str, pd.DataFrame], pd.DataFrame
        ] = lambda measure, results: results,
        to_observe: Callable[[Event], bool] = lambda event: True,
    ) -> None:
        """Registers an unstratified observation to the results system.

        Parameters
        ----------
        name
            Name of the observation. It will also be the name of the output results file
            for this particular observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        when
            Name of the lifecycle phase the observation should happen. Valid values are:
            "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
        requires_columns
            List of the state table columns that are required by either the `pop_filter` or the `aggregator`.
        requires_values
            List of the value pipelines that are required by either the `pop_filter` or the `aggregator`.
        results_gatherer
            Function that gathers the latest observation results.
        results_updater
            Function that updates existing raw observation results with newly gathered results.
        results_formatter
            Function that formats the raw observation results.
        to_observe
            Function that determines whether to perform an observation on this Event.

        Raises
        ------
        ValueError
            If any required callable arguments are missing.
        """
        required_callables = {
            "results_gatherer": results_gatherer,
            "results_updater": results_updater,
        }
        self._check_for_required_callables(name, required_callables)
        self._manager.register_observation(
            observation_type=UnstratifiedObservation,
            is_stratified=False,
            name=name,
            pop_filter=pop_filter,
            when=when,
            requires_columns=requires_columns,
            requires_values=requires_values,
            results_updater=results_updater,
            results_gatherer=results_gatherer,
            results_formatter=results_formatter,
            to_observe=to_observe,
        )

    def register_adding_observation(
        self,
        name: str,
        pop_filter: str = "tracked==True",
        when: str = "collect_metrics",
        requires_columns: List[str] = [],
        requires_values: List[str] = [],
        results_formatter: Callable[
            [str, pd.DataFrame], pd.DataFrame
        ] = lambda measure, results: results.reset_index(),
        additional_stratifications: List[str] = [],
        excluded_stratifications: List[str] = [],
        aggregator_sources: Optional[List[str]] = None,
        aggregator: Callable[[pd.DataFrame], Union[float, pd.Series]] = len,
        to_observe: Callable[[Event], bool] = lambda event: True,
    ) -> None:
        """Registers an adding observation to the results system.

        An "adding" observation is one that adds/sums new results to existing
        result values.

        Notes
        -----
        An adding observation is a specific type of stratified observation.

        Parameters
        ----------
        name
            Name of the observation. It will also be the name of the output results file
            for this particular observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        when
            Name of the lifecycle phase the observation should happen. Valid values are:
            "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
        requires_columns
            List of the state table columns that are required by either the `pop_filter` or the `aggregator`.
        requires_values
            List of the value pipelines that are required by either the `pop_filter` or the `aggregator`.
        results_formatter
            Function that formats the raw observation results.
        additional_stratifications
            List of additional :class:`Stratification <vivarium.framework.results.stratification.Stratification>`
            names by which to stratify this observation by.
        excluded_stratifications
            List of default :class:`Stratification <vivarium.framework.results.stratification.Stratification>`
            names to remove from this observation.
        aggregator_sources
            List of population view columns to be used in the `aggregator`.
        aggregator
            Function that computes the quantity for this observation.
        to_observe
            Function that determines whether to perform an observation on this Event.
        """
        self._manager.register_observation(
            observation_type=AddingObservation,
            is_stratified=True,
            name=name,
            pop_filter=pop_filter,
            when=when,
            requires_columns=requires_columns,
            requires_values=requires_values,
            results_formatter=results_formatter,
            additional_stratifications=additional_stratifications,
            excluded_stratifications=excluded_stratifications,
            aggregator_sources=aggregator_sources,
            aggregator=aggregator,
            to_observe=to_observe,
        )

    def register_concatenating_observation(
        self,
        name: str,
        pop_filter: str = "tracked==True",
        when: str = "collect_metrics",
        requires_columns: List[str] = [],
        requires_values: List[str] = [],
        results_formatter: Callable[
            [str, pd.DataFrame], pd.DataFrame
        ] = lambda measure, results: results,
        to_observe: Callable[[Event], bool] = lambda event: True,
    ) -> None:
        """Registers a concatenating observation to the results system.

        A "concatenating" observation is one that concatenates new results to
        existing results.

        Notes
        -----
        A concatenating observation is a specific type of unstratified observation.

        Parameters
        ----------
        name
            Name of the observation. It will also be the name of the output results file
            for this particular observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        when
            Name of the lifecycle phase the observation should happen. Valid values are:
            "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
        requires_columns
            List of the state table columns that are required by either the `pop_filter` or the `aggregator`.
        requires_values
            List of the value pipelines that are required by either the `pop_filter` or the `aggregator`.
        results_formatter
            Function that formats the raw observation results.
        to_observe
            Function that determines whether to perform an observation on this Event.
        """
        included_columns = ["event_time"] + requires_columns + requires_values
        self._manager.register_observation(
            observation_type=ConcatenatingObservation,
            is_stratified=False,
            name=name,
            pop_filter=pop_filter,
            when=when,
            requires_columns=requires_columns,
            requires_values=requires_values,
            results_formatter=results_formatter,
            included_columns=included_columns,
            to_observe=to_observe,
        )

    @staticmethod
    def _check_for_required_callables(
        observation_name: str, required_callables: Dict[str, Callable]
    ) -> None:
        """Raises a ValueError if any required callable arguments are missing."""
        missing = []
        for arg_name, callable in required_callables.items():
            if callable == _required_function_placeholder:
                missing.append(arg_name)
        if len(missing) > 0:
            raise ValueError(
                f"Observation '{observation_name}' is missing required callable(s): {missing}"
            )



from typing import Dict

import numpy as np
import pandas as pd
import pytest
from scipy import stats

from vivarium.framework.randomness import RESIDUAL_CHOICE, RandomnessError, RandomnessStream
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.randomness.stream import _normalize_shape, _set_residual_probability


@pytest.fixture
def randomness_stream():
    dates = [pd.Timestamp(1991, 1, 1), pd.Timestamp(1990, 1, 1)]
    randomness = RandomnessStream("test", dates.pop, 1, IndexMap())
    return randomness


def test_normalize_shape(weights_with_residuals, index):
    p = _normalize_shape(weights_with_residuals, index)
    assert p.shape == (len(index), len(weights_with_residuals))


def test__set_residual_probability(weights_with_residuals, index):
    # Coerce the weights to a 2-d numpy array.
    p = _normalize_shape(weights_with_residuals, index)

    residual = np.where(p == RESIDUAL_CHOICE, 1, 0)
    non_residual = np.where(p != RESIDUAL_CHOICE, p, 0)

    if np.any(non_residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received un-normalized probability weights.
            _set_residual_probability(p)

    elif np.any(residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received multiple instances of `RESIDUAL_CHOICE`
            _set_residual_probability(p)

    else:  # Things should work
        p_total = np.sum(_set_residual_probability(p))
        assert np.isclose(p_total, len(index), atol=0.0001)


def test_filter_for_probability_single_probability(randomness_stream, index):
    sub_index = randomness_stream.filter_for_probability(index, 0.5)
    assert np.isclose(len(sub_index) / len(index), 0.5, rtol=0.1)

    sub_sub_index = randomness_stream.filter_for_probability(sub_index, 0.5)
    assert np.isclose(len(sub_sub_index) / len(sub_index), 0.5, rtol=0.1)


def test_filter_for_probability_multiple_probabilities(randomness_stream, index):
    probabilities = pd.Series([0.3, 0.3, 0.3, 0.6, 0.6] * (index.size // 5), index=index)
    threshold_0_3 = probabilities.index[probabilities == 0.3]
    threshold_0_6 = probabilities.index.difference(threshold_0_3)

    sub_index = randomness_stream.filter_for_probability(index, probabilities)
    assert np.isclose(
        len(sub_index.intersection(threshold_0_3)) / len(threshold_0_3), 0.3, rtol=0.1
    )
    assert np.isclose(
        len(sub_index.intersection(threshold_0_6)) / len(threshold_0_6), 0.6, rtol=0.1
    )


def test_filter_for_rate_single_probability(randomness_stream, index):
    sub_index = randomness_stream.filter_for_rate(index, 0.5)
    assert np.isclose(len(sub_index) / len(index), 1 - np.exp(-0.5), rtol=0.1)

    sub_sub_index = randomness_stream.filter_for_rate(sub_index, 0.5)
    assert np.isclose(len(sub_sub_index) / len(sub_index), 1 - np.exp(-0.5), rtol=0.1)


def test_filter_for_rate_multiple_probabilities(randomness_stream, index):
    rates = pd.Series([0.3, 0.3, 0.3, 0.6, 0.6] * (index.size // 5), index=index)
    rate_0_3 = rates.index[rates == 0.3]
    rate_0_6 = rates.index.difference(rate_0_3)

    sub_index = randomness_stream.filter_for_rate(index, rates)
    assert np.isclose(
        len(sub_index.intersection(rate_0_3)) / len(rate_0_3), 1 - np.exp(-0.3), rtol=0.1
    )
    assert np.isclose(
        len(sub_index.intersection(rate_0_6)) / len(rate_0_6), 1 - np.exp(-0.6), rtol=0.1
    )


def test_choice(randomness_stream, index, choices, weights):
    chosen = randomness_stream.choice(index, choices, p=weights)
    count = chosen.value_counts()
    # If we have weights, normalize them, otherwise generate uniform weights.
    weights = (
        [w / sum(weights) for w in weights]
        if weights
        else [1 / len(choices) for _ in choices]
    )
    for k, c in count.items():
        assert np.isclose(c / len(index), weights[choices.index(k)], atol=0.01)


def test_choice_with_residuals(randomness_stream, index, choices, weights_with_residuals):
    print(RESIDUAL_CHOICE in weights_with_residuals)

    p = _normalize_shape(weights_with_residuals, index)

    residual = np.where(p == RESIDUAL_CHOICE, 1, 0)
    non_residual = np.where(p != RESIDUAL_CHOICE, p, 0)

    if np.any(non_residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received un-normalized probability weights.
            randomness_stream.choice(index, choices, p=weights_with_residuals)

    elif np.any(residual.sum(axis=1) > 1):
        with pytest.raises(RandomnessError):
            # We received multiple instances of `RESIDUAL_CHOICE`
            randomness_stream.choice(index, choices, p=weights_with_residuals)

    else:  # Things should work
        chosen = randomness_stream.choice(index, choices, p=weights_with_residuals)
        count = chosen.value_counts()
        print(weights_with_residuals)
        # We're relying on the fact that weights_with_residuals is a 1-d list
        residual_p = 1 - sum([w for w in weights_with_residuals if w != RESIDUAL_CHOICE])
        weights = [w if w != RESIDUAL_CHOICE else residual_p for w in weights_with_residuals]

        for k, c in count.items():
            assert np.isclose(c / len(index), weights[choices.index(k)], atol=0.01)


@pytest.mark.parametrize(
    "distribution, ppf, error_message",
    [
        (None, None, "Either distribution or ppf must be provided"),
        (stats.norm, lambda x: x, "Only one of distribution or ppf can be provided"),
    ],
)
def test_sample_from_distribution_bad_args(
    distribution, ppf, error_message, randomness_stream
):
    with pytest.raises(ValueError, match=error_message):
        randomness_stream.sample_from_distribution(
            index=pd.Index([]),
            distribution=distribution,
            ppf=ppf,
        )


@pytest.mark.parametrize(
    "distribution, params",
    [
        (stats.norm, {"loc": 5, "scale": 1}),
        (stats.beta, {"a": 1, "b": 2}),
    ],
)
def test_sample_from_distribution_using_scipy(
    index: pd.Index, distribution: stats.rv_continuous, params: Dict
):
    randomness_stream = RandomnessStream(
        "test", lambda: pd.Timestamp(2020, 1, 1), 1, IndexMap()
    )
    draws = randomness_stream.get_draw(index, "some_key")
    expected = distribution.ppf(draws, **params)

    sample = randomness_stream.sample_from_distribution(
        index=index, distribution=distribution, additional_key="some_key", **params
    )

    assert isinstance(sample, pd.Series)
    assert sample.index.equals(index)
    assert np.allclose(sample, expected)


def test_sample_from_distribution_using_ppf(index: pd.Index):
    def silly_ppf(x, add, mult):
        return mult * (x**2) + add

    randomness_stream = RandomnessStream(
        "test", lambda: pd.Timestamp(2020, 1, 1), 1, IndexMap()
    )
    draws = randomness_stream.get_draw(index, "some_key")
    expected = 2 * (draws**2) + 1

    sample = randomness_stream.sample_from_distribution(
        index=index, ppf=silly_ppf, additional_key="some_key", add=1, mult=2
    )

    assert isinstance(sample, pd.Series)
    assert sample.index.equals(index)
    assert np.allclose(sample, expected)



"""
=============
HDF Interface
=============

A convenience wrapper around the `tables <https://www.pytables.org>`_ and
:mod:`pandas` HDF interfaces.

Public Interface
----------------

The public interface consists of 5 functions:

.. list-table:: HDF Public Interface
   :widths: 20 60
   :header-rows: 1

   * - Function
     - Description
   * - :func:`touch`
     - Creates an HDF file, wiping an existing file if necessary.
   * - :func:`write`
     - Stores data at a key in an HDF file.
   * - :func:`load`
     - Loads (potentially filtered) data from a key in an HDF file.
   * - :func:`remove`
     - Clears data from a key in an HDF file.
   * - :func:`get_keys`
     - Gets all available HDF keys from an HDF file.

Contracts
+++++++++

- All functions in the public interface accept both :class:`pathlib.Path` and
  normal Python :class:`str` objects for paths.
- All functions in the public interface accept only :class:`str` objects
  as representations of the keys in the hdf file.  The strings must be
  formatted as ``"type.name.measure"`` or ``"type.measure"``.

"""

from __future__ import annotations

import json
import re
from pathlib import Path
from typing import Any, Literal

import pandas as pd
import tables  # type: ignore [import-untyped]
from tables.nodes import filenode  # type: ignore [import-untyped]

####################
# Public interface #
####################


def touch(path: Path | str) -> None:
    """Creates an HDF file, wiping an existing file if necessary.

    If the given path is proper to create a HDF file, it creates a new
    HDF file.

    Parameters
    ----------
    path
        The path to the HDF file.

    Raises
    ------
    ValueError
        If the non-proper path is given to create a HDF file.

    """
    path = _get_valid_hdf_path(path)

    with tables.open_file(str(path), mode="w"):
        pass


def write(path: Path | str, entity_key: str, data: Any) -> None:
    """Writes data to the HDF file at the given path to the given key.

    Parameters
    ----------
    path
        The path to the HDF file to write to.
    entity_key
        A string representation of the internal HDF path where we want to
        write the data. The key must be formatted as ``"type.name.measure"``
        or ``"type.measure"``.
    data
        The data to write. If it is a :mod:`pandas` object, it will be
        written using a
        `pandas.HDFStore <https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#hdf5-pytables>`_
        or :meth:`pandas.DataFrame.to_hdf`. If it is some other kind of python
        object, it will first be encoded as json with :func:`json.dumps` and
        then written to the provided key.

    Raises
    ------
    ValueError
        If the path or entity_key are improperly formatted.

    """
    hdf_path: Path = _get_valid_hdf_path(path)
    entity_key = EntityKey(entity_key)

    if isinstance(data, (pd.DataFrame, pd.Series)):
        _write_pandas_data(hdf_path, entity_key, data)
    else:
        _write_json_blob(hdf_path, entity_key, data)


def load(
    path: Path | str,
    entity_key: str,
    filter_terms: list[str] | None,
    column_filters: list[str] | None,
) -> Any:
    """Loads data from an HDF file.

    Parameters
    ----------
    path
        The path to the HDF file to load the data from.
    entity_key
        A representation of the internal HDF path where the data is located.
    filter_terms
        An optional list of terms used to filter the rows in the data.
        The terms must be formatted in a way that is suitable for use with
        the ``where`` argument of :func:`pandas.read_hdf`. Only
        filters applying to existing columns in the data are used.
    column_filters
        An optional list of columns to load from the data.

    Returns
    -------
        The data stored at the the given key in the HDF file.

    Raises
    ------
    ValueError
        If the path or entity_key are improperly formatted.
    """
    path = _get_valid_hdf_path(path)
    entity_key = EntityKey(entity_key)

    with tables.open_file(str(path)) as file:
        node = file.get_node(entity_key.path)
        if isinstance(node, tables.earray.EArray):
            # This should be a json encoded document rather than a pandas dataframe
            with filenode.open_node(node) as file_node:
                data = json.load(file_node)
        else:
            filter_terms = _get_valid_filter_terms(filter_terms, node.table.colnames)
            with pd.HDFStore(str(path), complevel=9, mode="r") as store:
                metadata = store.get_storer(  # type: ignore [operator]
                    entity_key.path
                ).attrs.metadata  # NOTE: must use attrs. write this up

            if metadata.get("is_empty", False):
                data = pd.read_hdf(path, entity_key.path, where=filter_terms)  # type: ignore [arg-type]
                data = data.set_index(
                    list(data.columns)
                )  # undoing transform performed on write
            else:
                data = pd.read_hdf(
                    path, entity_key.path, where=filter_terms, columns=column_filters  # type: ignore [arg-type]
                )

    return data


def remove(path: Path | str, entity_key: str) -> None:
    """Removes a piece of data from an HDF file.

    Parameters
    ----------
    path :
        The path to the HDF file to remove the data from.
    entity_key :
        A representation of the internal HDF path where the data is located.

    Raises
    ------
    ValueError
        If the path or entity_key are improperly formatted.
    """
    path = _get_valid_hdf_path(path)
    entity_key = EntityKey(entity_key)

    with tables.open_file(str(path), mode="a") as file:
        file.remove_node(entity_key.path, recursive=True)


def get_keys(path: Path | str) -> list[str]:
    """Gets key representation of all paths in an HDF file.

    Parameters
    ----------
    path
        The path to the HDF file.

    Returns
    -------
        A list of key representations of the internal paths in the HDF.
    """
    path = _get_valid_hdf_path(path)
    with tables.open_file(str(path)) as file:
        keys = _get_keys(file.root)
    return keys


class EntityKey(str):
    """A convenience wrapper that translates artifact keys.

    This class provides several representations of the artifact keys that
    are useful when working with the :mod:`pandas` and
    `tables <https://www.pytables.org>`_ HDF interfaces.

    """

    def __init__(self, key: str) -> None:
        """
        Parameters
        ----------
        key
            The string representation of the entity key.  Must be formatted
            as ``"type.name.measure"`` or ``"type.measure"``.

        Raises
        ------
        ValueError
            If the key is improperly formatted.
        """
        elements = [e for e in key.split(".") if e]
        if len(elements) not in [2, 3] or len(key.split(".")) != len(elements):
            raise ValueError(
                f"Invalid format for HDF key: {key}. "
                'Acceptable formats are "type.name.measure" and "type.measure"'
            )
        super().__init__()

    @property
    def type(self) -> str:
        """The type of the entity represented by the key."""
        return self.split(".")[0]

    @property
    def name(self) -> str:
        """The name of the entity represented by the key"""
        return self.split(".")[1] if len(self.split(".")) == 3 else ""

    @property
    def measure(self) -> str:
        """The measure associated with the data represented by the key."""
        return self.split(".")[-1]

    @property
    def group_prefix(self) -> str:
        """The HDF group prefix for the key."""
        return "/" + self.type if self.name else "/"

    @property
    def group_name(self) -> str:
        """The HDF group name for the key."""
        return self.name if self.name else self.type

    @property
    def group(self) -> str:
        """The full path to the group for this key."""
        return (
            self.group_prefix + "/" + self.group_name
            if self.name
            else self.group_prefix + self.group_name
        )

    @property
    def path(self) -> str:
        """The full HDF path associated with this key."""
        return self.group + "/" + self.measure

    def with_measure(self, measure: str) -> "EntityKey":
        """Replaces this key's measure with the provided one.

        Parameters
        ----------
        measure :
            The measure to replace this key's measure with.

        Returns
        -------
            A new EntityKey with the updated measure.
        """
        if self.name:
            return EntityKey(f"{self.type}.{self.name}.{measure}")
        else:
            return EntityKey(f"{self.type}.{measure}")

    def __eq__(self, other: object) -> bool:
        return isinstance(other, str) and str(self) == str(other)

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __hash__(self) -> int:
        return hash(str(self))

    def __repr__(self) -> str:
        return f"EntityKey({str(self)})"


#####################
# Private utilities #
#####################


def _get_valid_hdf_path(path: Path | str) -> Path:
    valid_suffixes = [".hdf", ".h5"]

    path = Path(path)
    if path.suffix not in valid_suffixes:
        raise ValueError(
            f"{str(path)} has an invalid HDF suffix {path.suffix}."
            f" HDF files must have one of {valid_suffixes} as a path suffix."
        )
    return path


def _write_pandas_data(
    path: Path, entity_key: EntityKey, data: pd.DataFrame | pd.Series[Any]
) -> None:
    """Write data in a pandas format to an HDF file.

    This method currently supports :class:`pandas DataFrame` objects, with or
    with or without columns, and :class:`pandas.Series` objects.
    """
    if data.empty:
        # Our data is indexed, sometimes with no other columns. This leaves an
        # empty dataframe that store.put will silently fail to write in table
        # format.
        data = data.reset_index()
        if data.empty:
            raise ValueError("Cannot write an empty dataframe that does not have an index.")
        metadata = {"is_empty": True}
        data_columns: Literal[True] | None = True
    else:
        metadata = {"is_empty": False}
        data_columns = None

    with pd.HDFStore(str(path), complevel=9) as store:
        store.put(entity_key.path, data, format="table", data_columns=data_columns)
        # NOTE: must use attrs. write this up
        store.get_storer(entity_key.path).attrs.metadata = metadata  # type: ignore [operator]


def _write_json_blob(path: Path, entity_key: EntityKey, data: Any) -> None:
    """Writes a Python object as json to the HDF file at the given path."""
    with tables.open_file(str(path), "a") as store:
        if entity_key.group_prefix not in store:
            store.create_group("/", entity_key.type)

        if entity_key.group not in store:
            store.create_group(entity_key.group_prefix, entity_key.group_name)

        with filenode.new_node(
            store, where=entity_key.group, name=entity_key.measure
        ) as fnode:
            fnode.write(bytes(json.dumps(data), "utf-8"))


def _get_keys(root: tables.node.Node, prefix: str = "") -> list[str]:
    """Recursively formats the paths in an HDF file into a key format."""
    keys = []
    for child in root:
        child_name = _get_node_name(child)
        if isinstance(child, tables.earray.EArray):  # This is the last node
            keys.append(f"{prefix}.{child_name}")
        elif isinstance(child, tables.table.Table):  # Parent was the last node
            keys.append(prefix)
        else:
            new_prefix = f"{prefix}.{child_name}" if prefix else child_name
            keys.extend(_get_keys(child, new_prefix))

    # Clean up some weird meta groups that get written with dataframes.
    keys = [k for k in keys if ".meta." not in k]
    return keys


def _get_node_name(node: tables.node.Node) -> str:
    """Gets the name of a node from its string representation."""
    node_string = str(node)
    node_path = node_string.split()[0]
    node_name = node_path.split("/")[-1]
    return node_name


def _get_valid_filter_terms(
    filter_terms: list[str] | None, colnames: list[str]
) -> list[str] | None:
    """Removes any filter terms referencing non-existent columns

    Parameters
    ----------
    filter_terms
        A list of terms formatted so as to be used in the `where` argument of
        :func:`pandas.read_hdf`.
    colnames :
        A list of column names present in the data that will be filtered.

    Returns
    -------
        The list of valid filter terms (terms that do not reference any column
        not existing in the data). Returns none if the list is empty because
        the `where` argument doesn't like empty lists.
    """
    if not filter_terms:
        return None
    valid_terms = filter_terms.copy()
    for term in filter_terms:
        # first strip out all the parentheses - the where in read_hdf
        # requires all references to be valid
        sub_term = re.sub("[()]", "", term)
        # then split each condition out
        split: list[str] = re.split("[&|]", sub_term)
        # get the unique columns referenced by this term
        term_columns = set([re.split(r"[<=>\s]", i.strip())[0] for i in split])
        if not term_columns.issubset(colnames):
            valid_terms.remove(term)
    return valid_terms if valid_terms else None



"""
===============
Stratifications
===============

"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable

import pandas as pd
from pandas.api.types import CategoricalDtype

STRATIFICATION_COLUMN_SUFFIX: str = "mapped_values"

# TODO: Parameterizing pandas objects fails below python 3.12
VectorMapper = Callable[[pd.DataFrame], pd.Series]  # type: ignore [type-arg]
ScalarMapper = Callable[[pd.Series], str]  # type: ignore [type-arg]


@dataclass
class Stratification:
    """Class for stratifying observed quantities by specified characteristics.
    Each Stratification represents a set of mutually exclusive and collectively
    exhaustive categories into which simulants can be assigned.
    This class includes a :meth:`stratify <stratify>` method that produces an
    output column by calling the mapper on the source columns.
    """

    name: str
    """Name of the stratification."""
    sources: list[str]
    """A list of the columns and values needed as input for the `mapper`."""
    categories: list[str]
    """Exhaustive list of all possible stratification values."""
    excluded_categories: list[str]
    """List of possible stratification values to exclude from results processing.
    If None (the default), will use exclusions as defined in the configuration."""
    mapper: VectorMapper | ScalarMapper | None
    """A callable that maps the columns and value pipelines specified by the
    `requires_columns` and `requires_values` arguments to the stratification
    categories. It can either map the entire population or an individual
    simulant. A simulation will fail if the `mapper` ever produces an invalid
    value."""
    is_vectorized: bool = False
    """True if the `mapper` function will map the entire population, and False
    if it will only map a single simulant."""

    def __str__(self) -> str:
        return (
            f"Stratification '{self.name}' with sources {self.sources}, "
            f"categories {self.categories}, and mapper {getattr(self.mapper, '__name__', repr(self.mapper))}"
        )

    def __post_init__(self) -> None:
        """Assign a default `mapper` if none was provided and check for non-empty
        `categories` and `sources` otherwise.
        Raises
        ------
        ValueError
            If no mapper is provided and the number of sources is not 1.
        ValueError
            If the categories argument is empty.
        ValueError
            If the sources argument is empty.
        """
        self.vectorized_mapper = self._get_vectorized_mapper(self.mapper, self.is_vectorized)
        if not self.categories:
            raise ValueError("The categories argument must be non-empty.")
        if not self.sources:
            raise ValueError("The sources argument must be non-empty.")

    def stratify(self, population: pd.DataFrame) -> pd.Series[CategoricalDtype]:
        """Apply the `mapper` to the population `sources` columns to create a new
        Series to be added to the population.

        Any `excluded_categories` (which have already been removed from `categories`)
        will be converted to NaNs in the new column and dropped later at the
        observation level.

        Parameters
        ----------
        population
            A DataFrame containing the data to be stratified.

        Returns
        -------
            A Series containing the mapped values to be used for stratifying.

        Raises
        ------
        ValueError
            If the mapper returns any values not in `categories` or `excluded_categories`.
        """
        mapped_column = self.vectorized_mapper(population[self.sources])
        unknown_categories = set(mapped_column) - set(
            self.categories + self.excluded_categories
        )
        # Reduce all nans to a single one
        unknown_categories = {cat for cat in unknown_categories if not pd.isna(cat)}
        if mapped_column.isna().any():
            unknown_categories.add(mapped_column[mapped_column.isna()].iat[0])
        if unknown_categories:
            raise ValueError(f"Invalid values mapped to {self.name}: {unknown_categories}")

        # Convert the dtype to the allowed categories. Note that this will
        # result in Nans for any values in excluded_categories.
        return mapped_column.astype(
            CategoricalDtype(categories=self.categories, ordered=True)
        )

    def _get_vectorized_mapper(
        self,
        user_provided_mapper: VectorMapper | ScalarMapper | None,
        is_vectorized: bool,
    ) -> VectorMapper:
        """
        Choose a VectorMapper based on the inputted callable mapper.
        """
        if user_provided_mapper is None:
            if len(self.sources) != 1:
                raise ValueError(
                    f"No mapper but {len(self.sources)} stratification sources are "
                    f"provided for stratification {self.name}. The list of sources "
                    "must be of length 1 if no mapper is provided."
                )
            return self._default_mapper
        elif is_vectorized:
            return user_provided_mapper  # type: ignore [return-value]
        else:
            return lambda population: population.apply(user_provided_mapper, axis=1)

    @staticmethod
    def _default_mapper(pop: pd.DataFrame) -> pd.Series[Any]:
        """Default stratification mapper that squeezes a DataFrame to a Series.

        Parameters
        ----------
        pop
            The data to be stratified.

        Returns
        -------
           The squeezed data to be stratified.

        Notes
        -----
        The input DataFrame is guaranteed to have a single column.
        """
        squeezed_pop: pd.Series[Any] = pop.squeeze(axis=1)
        return squeezed_pop


def get_mapped_col_name(col_name: str) -> str:
    """Return a new column name to be used for mapped values"""
    return f"{col_name}_{STRATIFICATION_COLUMN_SUFFIX}"


def get_original_col_name(col_name: str) -> str:
    """Return the original column name given a modified mapped column name."""
    return (
        col_name[: -(len(STRATIFICATION_COLUMN_SUFFIX)) - 1]
        if col_name.endswith(f"_{STRATIFICATION_COLUMN_SUFFIX}")
        else col_name
    )



"""
============================
The Plugin Management System
============================

.. todo::
   This part will come in with the full description of the plugin system
   in the next PR. -J.C. 05/07/19

"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Union

from layered_config_tree.main import LayeredConfigTree

from vivarium.exceptions import VivariumError
from vivarium.framework.utilities import import_by_path
from vivarium.manager import Interface, Manager

_MANAGERS = {
    "logging": {
        "controller": "vivarium.framework.logging.LoggingManager",
        "builder_interface": "vivarium.framework.logging.LoggingInterface",
    },
    "lookup": {
        "controller": "vivarium.framework.lookup.LookupTableManager",
        "builder_interface": "vivarium.framework.lookup.LookupTableInterface",
    },
    "randomness": {
        "controller": "vivarium.framework.randomness.RandomnessManager",
        "builder_interface": "vivarium.framework.randomness.RandomnessInterface",
    },
    "value": {
        "controller": "vivarium.framework.values.ValuesManager",
        "builder_interface": "vivarium.framework.values.ValuesInterface",
    },
    "event": {
        "controller": "vivarium.framework.event.EventManager",
        "builder_interface": "vivarium.framework.event.EventInterface",
    },
    "population": {
        "controller": "vivarium.framework.population.PopulationManager",
        "builder_interface": "vivarium.framework.population.PopulationInterface",
    },
    "resource": {
        "controller": "vivarium.framework.resource.ResourceManager",
        "builder_interface": "vivarium.framework.resource.ResourceInterface",
    },
}
DEFAULT_PLUGINS = {
    "plugins": {
        "required": {
            "component_manager": {
                "controller": "vivarium.framework.components.ComponentManager",
                "builder_interface": "vivarium.framework.components.ComponentInterface",
            },
            "clock": {
                "controller": "vivarium.framework.time.DateTimeClock",
                "builder_interface": "vivarium.framework.time.TimeInterface",
            },
            "component_configuration_parser": {
                "controller": "vivarium.framework.components.ComponentConfigurationParser",
                "builder_interface": None,
            },
            "lifecycle": {
                "controller": "vivarium.framework.lifecycle.LifeCycleManager",
                "builder_interface": "vivarium.framework.lifecycle.LifeCycleInterface",
            },
            "data": {
                "controller": "vivarium.framework.artifact.ArtifactManager",
                "builder_interface": "vivarium.framework.artifact.ArtifactInterface",
            },
            "results": {
                "controller": "vivarium.framework.results.ResultsManager",
                "builder_interface": "vivarium.framework.results.ResultsInterface",
            },
        },
        "optional": {},
    }
}


@dataclass
class PluginGroup:
    controller: Manager
    builder_interface: Interface | None


class PluginConfigurationError(VivariumError):
    """Error raised when plugin configuration is incorrectly specified."""

    pass


class PluginManager(Manager):
    def __init__(
        self,
        plugin_configuration: (
            dict[str, dict[str, dict[str, str]]] | LayeredConfigTree | None
        ) = None,
    ):
        self._plugin_configuration = LayeredConfigTree(
            DEFAULT_PLUGINS["plugins"], layers=["base", "override"]  # type: ignore [arg-type]
        )
        self._plugin_configuration.update(plugin_configuration, source="initialization_args")
        self._plugins: dict[str, PluginGroup] = {}

    def get_plugin(self, name: str) -> Manager:
        if name not in self._plugins:
            self._plugins[name] = self._get(name)
        return self._plugins[name].controller

    def get_plugin_interface(self, name: str) -> Interface | None:
        if name not in self._plugins:
            self._plugins[name] = self._get(name)
        return self._plugins[name].builder_interface

    def get_core_controllers(self) -> dict[str, Manager]:
        core_components = [
            name for name in self._plugin_configuration["required"].keys()  # type: ignore [union-attr]
        ] + list(_MANAGERS.keys())
        return {name: self.get_plugin(name) for name in core_components}

    def get_core_interfaces(self) -> dict[str, Interface | None]:
        core_components = [
            name for name in self._plugin_configuration["required"].keys()  # type: ignore [union-attr]
        ] + list(_MANAGERS.keys())
        return {name: self.get_plugin_interface(name) for name in core_components}

    def get_optional_controllers(self) -> dict[str, Manager]:
        return {
            name: self.get_plugin(name)
            for name in self._plugin_configuration["optional"].keys()  # type: ignore [union-attr]
        }

    def get_optional_interfaces(self) -> dict[str, Interface | None]:
        return {
            name: self.get_plugin_interface(name)
            for name in self._plugin_configuration["optional"].keys()  # type: ignore [union-attr]
        }

    def _get(self, name: str) -> PluginGroup:
        if name not in self._plugins:
            self._plugins[name] = self._build_plugin(name)
        return self._plugins[name]

    def _build_plugin(self, name: str) -> PluginGroup:
        plugin = self._lookup(name)

        try:
            controller = import_by_path(plugin["controller"])()
        except ValueError:
            raise PluginConfigurationError(
                f'Invalid plugin specification {plugin["controller"]}'
            )

        if plugin["builder_interface"] is not None:
            try:
                interface = import_by_path(plugin["builder_interface"])(controller)
            except ValueError:
                raise PluginConfigurationError(
                    f'Invalid plugin specification {plugin["builder_interface"]}'
                )
        else:
            interface = None

        return PluginGroup(controller=controller, builder_interface=interface)

    def _lookup(self, name: str) -> dict[str, str]:
        if name in self._plugin_configuration["required"]:  # type: ignore [operator]
            return self._plugin_configuration["required"][name]  # type: ignore [call-overload, index, return-value]
        elif name in self._plugin_configuration["optional"]:  # type: ignore [operator]
            return self._plugin_configuration["optional"][name]  # type: ignore [call-overload, index, return-value]
        elif name in _MANAGERS:
            return _MANAGERS[name]
        else:
            raise PluginConfigurationError(f"Plugin {name} not found.")

    def __repr__(self) -> str:
        return "PluginManager()"



# mypy: ignore-errors
"""
============================
The Vivarium Event Framework
============================

``vivarium`` constructs and manages the flow of :ref:`time <time_concept>`
through the emission of regularly scheduled events. The tools in this module
manage the relationships between event emitters and listeners and provide
an interface for user :ref:`components <components_concept>` to register
themselves as emitters or listeners to particular events.

The :class:`EventManager` maintains a mapping between event types and channels.
Each event type (and event types must be unique so event type is equivalent to
event name, e.g., ``time_step_prepare``) corresponds to an
:class:`EventChannel`, which tracks listeners to that event in prioritized
levels and passes on the event to those listeners when emitted.

The :class:`EventInterface` is exposed off the :ref:`builder <builder_concept>`
and provides two methods: :func:`get_emitter <EventInterface.get_emitter>`,
which returns a callable emitter for the given event type and
:func:`register_listener <EventInterface.register_listener>`, which adds the
given listener to the event channel for the given event. This is the only part
of the event framework with which client code should interact.

For more information, see the associated event
:ref:`concept note <event_concept>`.

"""

from typing import Any, Callable, Dict, List, NamedTuple, Optional

import pandas as pd

from vivarium.framework.lifecycle import ConstraintError
from vivarium.manager import Interface, Manager
from vivarium.types import ClockStepSize, ClockTime


class Event(NamedTuple):
    """An Event object represents the context of an event.

    Events themselves are just a bundle of data.  They must be emitted
    along an :class:`EventChannel` in order for other simulation components
    to respond to them.

    """

    #: An index into the population table containing all simulants affected
    #: by this event.
    index: pd.Index
    #: Any additional data provided by the user about the event.
    user_data: Dict[str, Any]
    #: The simulation time at which this event will resolve. The current
    #: simulation size plus the current time step size.
    time: ClockTime
    #: The current step size at the time of the event.
    step_size: ClockStepSize

    def split(self, new_index: pd.Index) -> "Event":
        """Create a copy of this event with a new index.

        This function should be used to emit an event in a new
        :class:`EventChannel` in response to an event emitted from a
        different channel.

        Parameters
        ----------
        new_index
            An index into the population table containing all simulants
            affected by this event.

        Returns
        -------
            The new event.
        """
        return Event(new_index, self.user_data, self.time, self.step_size)

    def __repr__(self):
        return (
            f"Event(user_data={self.user_data}, time={self.time}, step_size={self.step_size})"
        )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class EventChannel:
    """A named subscription channel that passes events to event listeners."""

    def __init__(self, manager, name):
        self.name = f"event_channel_{name}"
        self.manager = manager
        self.listeners = [[] for _ in range(10)]

    def emit(self, index: pd.Index, user_data: Optional[Dict] = None) -> Event:
        """Notifies all listeners to this channel that an event has occurred.

        Events are emitted to listeners in order of priority (with order 0 being
        first and order 9 last), with no ordering within a particular priority
        level guaranteed.

        Parameters
        ----------
        index
            An index into the population table containing all simulants
            affected by this event.
        user_data
            Any additional data provided by the user about the event.
        """
        if not user_data:
            user_data = {}
        e = Event(
            index,
            user_data,
            self.manager.clock() + self.manager.step_size(),
            self.manager.step_size(),
        )

        for priority_bucket in self.listeners:
            for listener in priority_bucket:
                listener(e)
        return e

    def __repr__(self):
        return f"EventChannel(listeners: {[listener for bucket in self.listeners for listener in bucket]})"


class EventManager(Manager):
    """The configuration for the event system.

    Notes
    -----
    Client code should never need to interact with this class
    except through the decorators in this module and the emitter
    function exposed on the builder during the setup phase.

    """

    def __init__(self):
        self._event_types = {}

    @property
    def name(self):
        """The name of this component."""
        return "event_manager"

    def get_channel(self, name):
        if name not in self._event_types:
            self._event_types[name] = EventChannel(self, name)
        return self._event_types[name]

    def setup(self, builder):
        """Performs this component's simulation setup.

        Parameters
        ----------
        builder
            Object giving access to core framework functionality.
        """
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()

        builder.event.register_listener("post_setup", self.on_post_setup)
        self.add_handlers = builder.lifecycle.add_handlers
        self.add_constraint = builder.lifecycle.add_constraint

        builder.lifecycle.add_constraint(
            self.get_emitter, allow_during=["setup", "simulation_end", "report"]
        )
        builder.lifecycle.add_constraint(self.register_listener, allow_during=["setup"])

    def on_post_setup(self, event):
        for name, channel in self._event_types.items():
            self.add_handlers(name, [h for level in channel.listeners for h in level])

    def get_emitter(self, name: str) -> Callable[[pd.Index, Optional[Dict]], Event]:
        """Get an emitter function for the named event.

        Parameters
        ----------
        name
            The name of the event.

        Returns
        -------
            A function that accepts an index and optional user data. This function
            creates and timestamps an Event and distributes it to all interested
            listeners
        """
        channel = self.get_channel(name)
        try:
            self.add_constraint(channel.emit, allow_during=[name])
        except ConstraintError:
            # Multiple components have requested this emitter.
            # Shouldn't happen in production, but happens frequently in tests.
            pass
        return channel.emit

    def register_listener(self, name: str, listener: Callable, priority: int = 5) -> None:
        """Registers a new listener to the named event.

        Parameters
        ----------
        name
            The name of the event.
        listener
            The consumer of the named event.
        priority
            Number in range(10) used to assign the ordering in which listeners
            process the event.
        """
        self.get_channel(name).listeners[priority].append(listener)

    def get_listeners(self, name: str) -> Dict[int, List[Callable]]:
        """Get all listeners registered for the named event.

        Parameters
        ----------
        name
            The name of the event.

        Returns
        -------
            A dictionary that maps each priority level of the named event's
            listeners to a list of listeners at that level.
        """
        channel = self.get_channel(name)
        return {
            priority: listeners
            for priority, listeners in enumerate(channel.listeners)
            if listeners
        }

    def list_events(self) -> List[Event]:
        """List all event names known to the event system.

        Returns
        -------
            A list of all known events.

        Notes
        -----
        This value can change after setup if components dynamically create
        new event labels.
        """
        return list(self._event_types.keys())

    def __contains__(self, item):
        return item in self._event_types

    def __repr__(self):
        return "EventManager()"


class EventInterface(Interface):
    """The public interface for the event system."""

    def __init__(self, manager: EventManager):
        self._manager = manager

    def get_emitter(self, name: str) -> Callable[[pd.Index, Optional[Dict]], Event]:
        """Gets an emitter for a named event.

        Parameters
        ----------
        name
            The name of the event the requested emitter will emit.
            Users may provide their own named events by requesting an emitter
            with this function, but should do so with caution as it makes time
            much more difficult to think about.

        Returns
        -------
            An emitter for the named event. The emitter should be called by
            the requesting component at the appropriate point in the simulation
            lifecycle.
        """
        return self._manager.get_emitter(name)

    def register_listener(
        self, name: str, listener: Callable[[Event], None], priority: int = 5
    ) -> None:
        """Registers a callable as a listener to a events with the given name.

        The listening callable will be called with a named ``Event`` as its
        only argument any time the event emitter is invoked from somewhere in
        the simulation.

        The framework creates the following events and emits them at different
        points in the simulation:

            - At the end of the setup phase: ``post_setup``
            - Every time step:
              - ``time_step__prepare``
              - ``time_step``
              - ``time_step__cleanup``
              - ``collect_metrics``
            - At simulation end: ``simulation_end``

        Parameters
        ----------
        name
            The name of the event to listen for.
        listener
            The callable to be invoked any time an :class:`Event` with the given
            name is emitted.
        priority
            One of {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}.
            An indication of the order in which event listeners should be
            called. Listeners with smaller priority values will be called
            earlier. Listeners with the same priority have no guaranteed
            ordering.  This feature should be avoided if possible. Components
            should strive to obey the Markov property as they transform the
            state table (the state of the simulation at the beginning of the
            next time step should only depend on the current state of the
            system).
        """
        self._manager.register_listener(name, listener, priority)



# mypy: ignore-errors
"""
===============
Results Context
===============

"""

from collections import defaultdict
from typing import Callable, Generator, List, Optional, Tuple, Type, Union

import pandas as pd
from pandas.core.groupby.generic import DataFrameGroupBy

from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.results.exceptions import ResultsConfigurationError
from vivarium.framework.results.observation import BaseObservation
from vivarium.framework.results.stratification import (
    Stratification,
    get_mapped_col_name,
    get_original_col_name,
)
from vivarium.types import ScalarValue


class ResultsContext:
    """Manager for organizing observations and their required stratifications.

    This context object is wholly contained by :class:`ResultsManager <vivarium.framework.results.manager.ResultsManager>`.
    Stratifications and observations can be added to the context through the manager via the
    :meth:`add_stratification <vivarium.framework.results.context.ResultsContext.add_stratification>` and
    :meth:`register_observation <vivarium.framework.results.context.ResultsContext.register_observation>` methods, respectively.

    Attributes
    ----------
    default_stratifications
        List of column names to use for stratifying results.
    stratifications
        List of :class:`Stratification <vivarium.framework.results.stratification.Stratification>`
        objects to be applied to results.
    excluded_categories
        Dictionary of possible per-metric stratification values to be excluded
        from results processing.
    observations
        Dictionary of observation details. It is of the format
        {lifecycle_phase: {(pop_filter, stratifications): list[Observation]}}.
        Allowable lifecycle_phases are "time_step__prepare", "time_step",
        "time_step__cleanup", and "collect_metrics".
    logger
        Logger for the results context.
    """

    def __init__(self) -> None:
        self.default_stratifications: List[str] = []
        self.stratifications: List[Stratification] = []
        self.excluded_categories: dict[str, list[str]] = {}
        self.observations: defaultdict = defaultdict(lambda: defaultdict(list))

    @property
    def name(self) -> str:
        return "results_context"

    def setup(self, builder: Builder) -> None:
        """Set up the results context.

        This method is called by the :class:`ResultsManager <vivarium.framework.results.manager.ResultsManager>`
        during the setup phase of that object.
        """
        self.logger = builder.logging.get_logger(self.name)
        self.excluded_categories = (
            builder.configuration.stratification.excluded_categories.to_dict()
        )

    # noinspection PyAttributeOutsideInit
    def set_default_stratifications(self, default_grouping_columns: List[str]) -> None:
        """Set the default stratifications to be used by stratified observations.

        Parameters
        ----------
        default_grouping_columns
            List of stratifications to be used.

        Raises
        ------
        ResultsConfigurationError
            If the `self.default_stratifications` attribute has already been set.
        """
        if self.default_stratifications:
            raise ResultsConfigurationError(
                "Multiple calls are being made to set default grouping columns "
                "for results production."
            )
        self.default_stratifications = default_grouping_columns

    def add_stratification(
        self,
        name: str,
        sources: List[str],
        categories: List[str],
        excluded_categories: Optional[List[str]],
        mapper: Optional[
            Union[
                Callable[[Union[pd.Series, pd.DataFrame]], pd.Series],
                Callable[[ScalarValue], str],
            ]
        ],
        is_vectorized: bool,
    ) -> None:
        """Add a stratification to the results context.

        Parameters
        ----------
        name
            Name of the stratification.
        sources
            A list of the columns and values needed as input for the `mapper`.
        categories
            Exhaustive list of all possible stratification values.
        excluded_categories
            List of possible stratification values to exclude from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        mapper
            A callable that maps the columns and value pipelines specified by
            `sources` to the stratification categories. It can either map the entire
            population or an individual simulant. A simulation will fail if the `mapper`
            ever produces an invalid value.
        is_vectorized
            True if the `mapper` function will map the entire population, and False
            if it will only map a single simulant.

        Raises
        ------
        ValueError
            If the stratification `name` is already used.
        ValueError
            If there are duplicate `categories`.
        ValueError
            If any `excluded_categories` are not in `categories`.
        """
        already_used = [
            stratification
            for stratification in self.stratifications
            if stratification.name == name
        ]
        if already_used:
            raise ValueError(
                f"Stratification name '{name}' is already used: {str(already_used[0])}."
            )
        unique_categories = set(categories)
        if len(categories) != len(unique_categories):
            for category in unique_categories:
                categories.remove(category)
            raise ValueError(
                f"Found duplicate categories in stratification '{name}': {categories}."
            )

        # Handle excluded categories. If excluded_categories are explicitly
        # passed in, we use that instead of what is in the model spec.
        to_exclude = (
            excluded_categories
            if excluded_categories is not None
            else self.excluded_categories.get(name, [])
        )
        unknown_exclusions = set(to_exclude) - set(categories)
        if len(unknown_exclusions) > 0:
            raise ValueError(
                f"Excluded categories {unknown_exclusions} not found in categories "
                f"{categories} for stratification '{name}'."
            )
        if to_exclude:
            self.logger.debug(
                f"'{name}' has category exclusion requests: {to_exclude}\n"
                "Removing these from the allowable categories."
            )
            categories = [category for category in categories if category not in to_exclude]

        stratification = Stratification(
            name,
            sources,
            categories,
            to_exclude,
            mapper,
            is_vectorized,
        )
        self.stratifications.append(stratification)

    def register_observation(
        self,
        observation_type: Type[BaseObservation],
        name: str,
        pop_filter: str,
        when: str,
        **kwargs,
    ) -> None:
        """Add an observation to the results context.

        Parameters
        ----------
        observation_type
            Specific class type of observation to register.
        name
            Name of the observation. It will also be the name of the output results file
            for this particular observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        when
            Name of the lifecycle phase the observation should happen. Valid values are:
            "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
        **kwargs
            Additional keyword arguments to be passed to the observation's constructor.

        Raises
        ------
        ValueError
            If the observation `name` is already used.
        """
        already_used = None
        if self.observations:
            # NOTE: self.observations is a list where each item is a dictionary
            # of the form {lifecycle_phase: {(pop_filter, stratifications): List[Observation]}}.
            # We use a triple-nested for loop to iterate over only the list of Observations
            # (i.e. we do not need the lifecycle_phase, pop_filter, or stratifications).
            for observation_details in self.observations.values():
                for observations in observation_details.values():
                    for observation in observations:
                        if observation.name == name:
                            already_used = observation
        if already_used:
            raise ValueError(
                f"Observation name '{name}' is already used: {str(already_used)}."
            )

        # Instantiate the observation and add it and its (pop_filter, stratifications)
        # tuple as a key-value pair to the self.observations[when] dictionary.
        observation = observation_type(name=name, pop_filter=pop_filter, when=when, **kwargs)
        self.observations[observation.when][
            (observation.pop_filter, observation.stratifications)
        ].append(observation)

    def gather_results(
        self, population: pd.DataFrame, lifecycle_phase: str, event: Event
    ) -> Generator[
        Tuple[
            Optional[pd.DataFrame],
            Optional[str],
            Optional[Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame]],
        ],
        None,
        None,
    ]:
        """Generate and yield current results for all observations at this lifecycle
        phase and event.

        Each set of results are stratified and grouped by
        all registered stratifications as well as filtered by their respective
        observation's pop_filter.

        Parameters
        ----------
        population
            The current population DataFrame.
        lifecycle_phase
            The current lifecycle phase.
        event
            The current Event.

        Yields
        ------
            A tuple containing each observation's newly observed results, the name of
            the observation, and the observations results updater function. Note that
            it yields (None, None, None) if the filtered population is empty.

        Raises
        ------
        ValueError
            If a stratification's temporary column name already exists in the population DataFrame.
        """

        for stratification in self.stratifications:
            # Add new columns of mapped values to the population to prevent name collisions
            new_column = get_mapped_col_name(stratification.name)
            if new_column in population.columns:
                raise ValueError(
                    f"Stratification column '{new_column}' "
                    "already exists in the state table or as a pipeline which is a required "
                    "name for stratifying results - choose a different name."
                )
            population[new_column] = stratification.stratify(population)

        # Optimization: We store all the producers by pop_filter and stratifications
        # so that we only have to apply them once each time we compute results.
        for (pop_filter, stratification_names), observations in self.observations[
            lifecycle_phase
        ].items():
            # Results production can be simplified to
            # filter -> groupby -> aggregate in all situations we've seen.
            filtered_pop = self._filter_population(
                population, pop_filter, stratification_names
            )
            if filtered_pop.empty:
                yield None, None, None
            else:
                if stratification_names is None:
                    pop = filtered_pop
                else:
                    pop = self._get_groups(stratification_names, filtered_pop)
                for observation in observations:
                    results = observation.observe(event, pop, stratification_names)
                    if results is not None:
                        self._rename_stratification_columns(results)

                    yield (results, observation.name, observation.results_updater)

    def _filter_population(
        self,
        population: pd.DataFrame,
        pop_filter: str,
        stratification_names: Optional[tuple[str, ...]],
    ) -> pd.DataFrame:
        """Filter out simulants not to observe."""
        pop = population.query(pop_filter) if pop_filter else population.copy()
        if stratification_names:
            # Drop all rows in the mapped_stratification columns that have NaN values
            # (which only exist if the mapper returned an excluded category).
            pop = pop.dropna(
                subset=[
                    get_mapped_col_name(stratification)
                    for stratification in stratification_names
                ]
            )
        return pop

    @staticmethod
    def _get_groups(
        stratifications: Tuple[str, ...], filtered_pop: pd.DataFrame
    ) -> DataFrameGroupBy:
        """Group the population by stratification.

        Notes
        -----
        Stratifications at this point can be an empty tuple.

        HACK: If there are no `stratifications` (i.e. it's an empty tuple), we
        create a single group of the entire `filtered_pop` index and assign
        it a name of "all". The alternative is to use the entire population
        instead of a groupby object, but then we would need to handle
        the different ways the aggregator can behave.
        """

        if stratifications:
            pop_groups = filtered_pop.groupby(
                [get_mapped_col_name(stratification) for stratification in stratifications],
                observed=False,
            )
        else:
            pop_groups = filtered_pop.groupby(lambda _: "all")
        return pop_groups

    def _rename_stratification_columns(self, results: pd.DataFrame) -> None:
        """Convert the temporary stratified mapped index names back to their original names."""
        if isinstance(results.index, pd.MultiIndex):
            idx_names = [get_original_col_name(name) for name in results.index.names]
            results.rename_axis(index=idx_names, inplace=True)
        else:
            idx_name = results.index.name
            if idx_name is not None:
                results.index.rename(get_original_col_name(idx_name), inplace=True)



# mypy: ignore-errors
"""
=================
The Data Artifact
=================

This module provides tools for interacting with data artifacts.

A data artifact is an archive on disk intended to package up all data
relevant to a particular simulation. This module provides a class to wrap that
archive file for convenient access and inspection.

"""

import re
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from vivarium.framework.artifact import hdf


class ArtifactException(Exception):
    """Exception raise for inconsistent use of the data artifact."""

    pass


class Artifact:
    """An interface for interacting with :mod:`vivarium` artifacts."""

    def __init__(self, path: Union[str, Path], filter_terms: Optional[List[str]] = None):
        """
        Parameters
        ----------
        path
            The path to the artifact file.
        filter_terms
            A set of terms suitable for usage with the ``where`` kwarg
            for :func:`pandas.read_hdf`.
        """
        self._path = Path(path)
        self._filter_terms = filter_terms
        self._draw_column_filter = _parse_draw_filters(filter_terms)
        self._cache = {}

        self.create_hdf_with_keyspace(self._path)
        self._keys = Keys(self._path)

    @property
    def path(self):
        """The path to the artifact file."""
        return str(self._path)

    @property
    def keys(self) -> List[str]:
        """A list of all the keys contained within the artifact."""
        return self._keys.to_list()

    @property
    def filter_terms(self) -> List[str]:
        """Filters that will be applied to the requested data on loads."""
        return self._filter_terms

    @staticmethod
    def create_hdf_with_keyspace(path: Path):
        """Creates the artifact HDF file and adds a node to track keys."""
        if not path.is_file():
            warnings.warn(f"No artifact found at {path}. Building new artifact.")
            hdf.touch(path)

        keys = hdf.get_keys(path)
        if keys and "metadata.keyspace" not in keys:
            raise ArtifactException(
                "Attempting to construct an Artifact from a malformed existing file. "
                "This can occur when constructing an Artifact from an existing file when "
                "the existing file was generated by some other hdf writing mechanism "
                "(e.g. pandas.to_hdf) rather than generating the the file using this class "
                "and a non-existent or empty hdf file."
            )
        if not keys:
            hdf.write(path, "metadata.keyspace", ["metadata.keyspace"])

    def load(self, entity_key: str) -> Any:
        """Loads the data associated with provided entity_key.

        Parameters
        ----------
        entity_key
            The key associated with the expected data.

        Returns
        -------
            The expected data. Will either be a standard Python object or a
            :class:`pandas.DataFrame` or :class:`pandas.Series`.

        Raises
        ------
        ArtifactException
            If the provided key is not in the artifact.
        """
        if entity_key not in self:
            raise ArtifactException(f"{entity_key} should be in {self.path}.")

        if entity_key not in self._cache:
            data = hdf.load(
                self._path, entity_key, self._filter_terms, self._draw_column_filter
            )
            # FIXME: Under what conditions do we get None here.
            assert (
                data is not None
            ), f"Data for {entity_key} is not available. Check your model specification."
            self._cache[entity_key] = data

        return self._cache[entity_key]

    def write(self, entity_key: str, data: Any) -> None:
        """Writes data into the artifact and binds it to the provided key.

        Parameters
        ----------
        entity_key
            The key associated with the provided data.
        data
            The data to write. Accepted formats are :class:`pandas.Series`,
            :class:`pandas.DataFrame` or standard python types and containers.

        Raises
        ------
        ArtifactException
            If the provided key already exists in the artifact.
        """
        if entity_key in self:
            raise ArtifactException(f"{entity_key} already in artifact.")
        elif data is None:
            raise ArtifactException(f"Attempting to write to key {entity_key} with no data.")
        else:
            hdf.write(self._path, entity_key, data)
            self._keys.append(entity_key)

    def remove(self, entity_key: str) -> None:
        """Removes data associated with the provided key from the artifact.

        Parameters
        ----------
        entity_key
            The key associated with the data to remove.

        Raises
        ------
        ArtifactException
            If the key is not present in the artifact.
        """
        if entity_key not in self:
            raise ArtifactException(
                f"Trying to remove non-existent key {entity_key} from artifact."
            )

        self._keys.remove(entity_key)
        if entity_key in self._cache:
            self._cache.pop(entity_key)
        hdf.remove(self._path, entity_key)

    def replace(self, entity_key: str, data: Any) -> None:
        """Replaces the artifact data at the provided key with the new data.

        Parameters
        ----------
        entity_key
            The key for which the data should be overwritten.
        data
            The data to write. Accepted formats are :class:`pandas.Series`,
            :class:`pandas.DataFrame` or standard python types and containers.

        Raises
        ------
        ArtifactException
            If the provided key does not already exist in the artifact.
        """
        if entity_key not in self:
            raise ArtifactException(
                f"Trying to replace non-existent key {entity_key} in artifact."
            )
        self.remove(entity_key)
        self.write(entity_key, data)

    def clear_cache(self) -> None:
        """Clears the artifact's cache.

        The artifact will cache data in memory to improve performance for
        repeat access.
        """
        self._cache = {}

    def __iter__(self):
        return iter(self.keys)

    def __contains__(self, item: str):
        return item in self.keys

    def __repr__(self):
        return f"Artifact(keys={self.keys})"

    def __str__(self):
        key_tree = _to_tree(self.keys)
        out = "Artifact containing the following keys:\n"
        for root, children in key_tree.items():
            out += f"{root}\n"
            for child, grandchildren in children.items():
                out += f"\t{child}\n"
                for grandchild in grandchildren:
                    out += f"\t\t{grandchild}\n"
        return out


def _to_tree(keys: List[str]) -> Dict[str, Dict[str, List[str]]]:
    out = defaultdict(lambda: defaultdict(list))
    for k in keys:
        key = k.split(".")
        if len(key) == 3:
            out[key[0]][key[1]].append(key[2])
        else:
            out[key[0]][key[1]] = []
    return out


class Keys:
    """A convenient wrapper around the keyspace which makes it easier for
    Artifact to maintain its keyspace when an entity key is added or removed.

    With the artifact_path, Keys object is initialized when the Artifact is
    initialized

    """

    keyspace_node = "metadata.keyspace"

    def __init__(self, artifact_path: Path):
        self._path = artifact_path
        self._keys = [str(k) for k in hdf.load(self._path, "metadata.keyspace", None, None)]

    def append(self, new_key: str) -> None:
        """Whenever the artifact gets a new key and new data, append is called to
        remove the old keyspace and to write the updated keyspace"""

        self._keys.append(new_key)
        hdf.remove(self._path, self.keyspace_node)
        hdf.write(self._path, self.keyspace_node, self._keys)

    def remove(self, removing_key: str) -> None:
        """Whenever the artifact removes a key and data, remove is called to
        remove the key from keyspace and write the updated keyspace."""

        self._keys.remove(removing_key)
        hdf.remove(self._path, self.keyspace_node)
        hdf.write(self._path, self.keyspace_node, self._keys)

    def to_list(self) -> List[str]:
        """A list of all the entity keys in the associated artifact."""

        return self._keys

    def __contains__(self, item):
        return item in self._keys


def _parse_draw_filters(filter_terms) -> Optional[list[str]]:
    """Given a list of filter terms, parse out any related to draws and convert
    to the list of column names.

    Also include 'value' column for compatibility with data that is long on draws.
    """
    columns = None

    if filter_terms:
        draw_terms = []
        for term in filter_terms:
            # first strip out all the parentheses
            t = re.sub("[()]", "", term)
            # then split each condition out
            t = re.split("[&|]", t)
            # then split condition to see if it relates to draws
            split_term = [re.split("([<=>in])", i) for i in t]
            draw_terms.extend([t for t in split_term if t[0].strip() == "draw"])

        if len(draw_terms) > 1:
            raise ValueError(
                f"You can only supply one filter term related to draws. "
                f"You supplied {filter_terms}, {len(draw_terms)} of which pertain to draws."
            )

        if draw_terms:
            # convert term to columns
            term = [s.strip() for s in draw_terms[0] if s.strip()]
            if len(term) == 4 and term[1].lower() == "i" and term[2].lower() == "n":
                draws = [int(d) for d in term[-1][1:-1].split(",")]
            elif (len(term) == 4 and term[1] == term[2] == "=") or (
                len(term) == 3 and term[1] == "="
            ):
                draws = [int(term[-1])]
            else:
                raise NotImplementedError(
                    f"The only supported draw filters are =, ==, or in. "
                    f'You supplied {"".join(term)}.'
                )

            columns = [f"draw_{n}" for n in draws] + ["value"]

    return columns



"""
==============================
Random Numbers in ``vivarium``
==============================

This module contains classes and functions supporting common random numbers.

Vivarium has some peculiar needs around randomness. We need to be totally
consistent between branches in a comparison. For example, if a simulant gets
hit by a truck in the base case it must be hit by that same truck in the
counter-factual at exactly the same moment unless the counter-factual
explicitly deals with traffic accidents. That means that the system can't rely
on standard global randomness sources because small changes to the number of
bits consumed or the order in which randomness consuming operations occur will
cause the system to diverge.

The current approach is to generate hash-based
seeds where the key is the simulation time, the simulant's id, the draw number
and a unique id for the decision point which needs the randomness.
These seeds are then used to generate `numpy.random.RandomState` objects that
can be used to create pseudo-random numbers in a repeatable manner.


For more information, see the Common Random Numbers
:ref:`concept note <crn_concept>`.

"""

from vivarium.framework.randomness.exceptions import RandomnessError
from vivarium.framework.randomness.manager import RandomnessInterface, RandomnessManager
from vivarium.framework.randomness.stream import RESIDUAL_CHOICE, RandomnessStream, get_hash



# mypy: ignore-errors
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event


class Mortality(Component):
    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        """A set of default configuration values for this component.

        These can be overwritten in the simulation model specification or by
        providing override values when constructing an interactive simulation.
        """
        return {
            "mortality": {
                "mortality_rate": 0.01,
            }
        }

    @property
    def columns_required(self) -> Optional[List[str]]:
        return ["tracked", "alive"]

    #####################
    # Lifecycle methods #
    #####################

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        """Performs this component's simulation setup.

        The ``setup`` method is automatically called by the simulation
        framework. The framework passes in a ``builder`` object which
        provides access to a variety of framework subsystems and metadata.

        Parameters
        ----------
        builder
            Access to simulation tools and subsystems.
        """
        self.config = builder.configuration.mortality
        self.randomness = builder.randomness.get_stream("mortality")

        self.mortality_rate = builder.value.register_rate_producer(
            "mortality_rate", source=self.base_mortality_rate
        )

    ########################
    # Event-driven methods #
    ########################

    def on_time_step(self, event: Event) -> None:
        """Determines who dies each time step.

        Parameters
        ----------
        event
            An event object emitted by the simulation containing an index
            representing the simulants affected by the event and timing
            information.
        """
        effective_rate = self.mortality_rate(event.index)
        effective_probability = 1 - np.exp(-effective_rate)
        draw = self.randomness.get_draw(event.index)
        affected_simulants = draw < effective_probability
        self.population_view.subview("alive").update(
            pd.Series("dead", index=event.index[affected_simulants])
        )

    def on_time_step_prepare(self, event: Event) -> None:
        """Untrack any simulants who died during the previous time step.

        We do this after the previous time step because the mortality
        observer needs to collect observations before updating.

        Parameters
        ----------
        event
            An event object emitted by the simulation containing an index
            representing the simulants affected by the event and timing
            information.
        """
        population = self.population_view.get(event.index)
        population.loc[
            (population["alive"] == "dead") & population["tracked"] == True, "tracked"
        ] = False
        self.population_view.update(population)

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def base_mortality_rate(self, index: pd.Index) -> pd.Series:
        """Computes the base mortality rate for every individual.

        Parameters
        ----------
        index
            A representation of the simulants to compute the base mortality
            rate for.

        Returns
        -------
            The base mortality rate for all simulants in the index.
        """
        return pd.Series(self.config.mortality_rate, index=index)



from unittest.mock import call

import pytest
import yaml
from layered_config_tree.main import LayeredConfigTree

from tests.helpers import MockComponentA, MockComponentB
from vivarium.framework.components.parser import ComponentConfigurationParser, ParsingError
from vivarium.framework.configuration import build_simulation_configuration

TEST_COMPONENTS_NESTED = """
components:
    ministry:
        silly_walk:
           - Prance()
           - Jump('front_flip')
           - PratFall('15')
    pet_shop:
       - Parrot()
       - dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
    news: SomethingCompletelyDifferent()
"""

# common case when users comment out and ends up having none
TEST_COMPONENTS_EMPTY_KEY = """
components:
    ministry.silly_walk:
       - Prance()
       - Jump('front_flip')
       - PratFall('15')
    pet_shop:
#       - Parrot()
#       - dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
    news: SomethingCompletelyDifferent()
"""

TEST_COMPONENTS_NON_STRING = """
components:
    ministry.silly_walk:
       - 42
       - Jump('front_flip')
       - PratFall('15')
    pet_shop:
#       - Parrot()
#       - dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
    news: SomethingCompletelyDifferent()
"""

TEST_COMPONENTS_FLAT = """
components:
    - ministry.silly_walk.Prance()
    - ministry.silly_walk.Jump('front_flip')
    - ministry.silly_walk.PratFall('15')
    - pet_shop.Parrot()
    - pet_shop.dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')
    - news.SomethingCompletelyDifferent()
"""

TEST_COMPONENTS_PARSED = [
    "ministry.silly_walk.Prance()",
    "ministry.silly_walk.Jump('front_flip')",
    "ministry.silly_walk.PratFall('15')",
    "pet_shop.Parrot()",
    "pet_shop.dangerous_animals.Crocodile('gold_tooth', 'teacup', '3.14')",
    "news.SomethingCompletelyDifferent()",
]

TEST_COMPONENTS_PREPPED = [
    ("ministry.silly_walk.Prance", tuple()),
    ("ministry.silly_walk.Jump", ("front_flip",)),
    ("ministry.silly_walk.PratFall", ("15",)),
    ("pet_shop.Parrot", tuple()),
    ("pet_shop.dangerous_animals.Crocodile", ("gold_tooth", "teacup", "3.14")),
    ("news.SomethingCompletelyDifferent", tuple()),
]


def mock_importer(path):
    return {
        "test_components.MockComponentA": MockComponentA,
        "test_components.MockComponentB": MockComponentB,
    }[path]


@pytest.fixture()
def parser(mocker) -> ComponentConfigurationParser:
    parser = ComponentConfigurationParser()
    parser.import_and_instantiate_component = mocker.Mock()
    return parser


@pytest.fixture(params=[TEST_COMPONENTS_NESTED, TEST_COMPONENTS_FLAT])
def components(request):
    return request.param


@pytest.fixture
def import_and_instantiate_mock(mocker):
    return mocker.patch(
        "vivarium.framework.components.parser.import_and_instantiate_component"
    )


def test_prep_component(parser):
    desc = 'cave_system.monsters.Rabbit("timid", "squeak")'
    component, args = parser.prep_component(desc)
    assert component == "cave_system.monsters.Rabbit"
    assert set(args) == {"timid", "squeak"}


def test_prep_component_syntax_error(parser):
    desc = 'cave_system.monsters.Rabbit("timid", 0.01)'
    with pytest.raises(ParsingError):
        parser.prep_component(desc)

    desc = 'cave_system.monsters.Rabbit("timid\', "0.01")'
    with pytest.raises(ParsingError):
        parser.prep_component(desc)

    desc = "cave_system.monsters.Rabbit(\"timid', '0.01')"
    with pytest.raises(ParsingError):
        parser.prep_component(desc)


def test_parse_and_prep_components(parser):
    prepped_components = [
        parser.prep_component(component) for component in TEST_COMPONENTS_PARSED
    ]

    assert set(TEST_COMPONENTS_PREPPED) == set(prepped_components)


def test_import_and_instantiate_components(monkeypatch):
    monkeypatch.setattr("vivarium.framework.components.parser.import_by_path", mock_importer)

    component_descriptions = [
        ("test_components.MockComponentA", ("A Hundred and One Ways to Start a Fight",)),
        ("test_components.MockComponentB", ("Ethel the Aardvark goes Quantity Surveying",)),
    ]

    parser = ComponentConfigurationParser()
    component_list = [
        parser.import_and_instantiate_component(path, arg_tuple)
        for path, arg_tuple in component_descriptions
    ]

    assert len(component_list) == 2
    assert isinstance(component_list[0], MockComponentA)
    assert component_list[0].args == ("A Hundred and One Ways to Start a Fight",)
    assert isinstance(component_list[1], MockComponentB)
    assert component_list[1].args == ("Ethel the Aardvark goes Quantity Surveying",)


def test_get_components(parser, components):
    config = build_simulation_configuration()
    config.update(components)

    parser.get_components(config.components)

    calls = [call(path, args) for path, args in TEST_COMPONENTS_PREPPED]
    parser.import_and_instantiate_component.assert_has_calls(calls)


@pytest.mark.parametrize(
    "config, error_message",
    [
        (TEST_COMPONENTS_EMPTY_KEY, "empty with the header"),
        (TEST_COMPONENTS_NON_STRING, "should be a string, list, or dictionary"),
    ],
)
def test_components_invalid_config(parser, config, error_message):
    bad_config = LayeredConfigTree(yaml.full_load(config))["components"]
    with pytest.raises(ParsingError, match=error_message):
        parser.parse_component_config(bad_config)



import json
import random
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import tables
from tables.nodes import filenode

from vivarium.framework.artifact import hdf
from vivarium.testing_utilities import build_table

_KEYS = [
    "population.age_bins",
    "population.structure",
    "population.theoretical_minimum_risk_life_expectancy",
    "cause.all_causes.restrictions",
    "metadata.versions",
    "metadata.locations",
    "metadata.keyspace",
]


@pytest.fixture
def hdf_keys():
    return _KEYS


@pytest.fixture(params=_KEYS)
def hdf_key(request):
    return request.param


@pytest.fixture(
    params=[
        "totally.new.thing",
        "other.new_thing",
        "cause.sort_of_new",
        "cause.also.new",
        "cause.all_cause.kind_of_new",
    ]
)
def mock_key(request):
    return hdf.EntityKey(request.param)


@pytest.fixture(params=[[], {}, ["data"], {"thing": "value"}, "bananas"])
def json_data(request):
    return request.param


def test_touch_no_file(mocker):
    path = Path("not/an/existing/path.hdf")
    tables_mock = mocker.patch("vivarium.framework.artifact.hdf.tables")

    hdf.touch(path)
    tables_mock.open_file.assert_called_once_with(str(path), mode="w")
    tables_mock.reset_mock()


def test_touch_exists_but_not_hdf_file_path(hdf_file_path):
    dir_path = Path(hdf_file_path).parent
    with pytest.raises(ValueError):
        hdf.touch(dir_path)
    non_hdf_path = Path(hdf_file_path).parent / "test.txt"
    with pytest.raises(ValueError):
        hdf.touch(non_hdf_path)


def test_touch_existing_file(tmpdir):
    path = f"{str(tmpdir)}/test.hdf"

    hdf.touch(path)
    hdf.write(path, hdf.EntityKey("test.key"), "data")
    assert hdf.get_keys(path) == ["test.key"]

    # should wipe out and make it again
    hdf.touch(path)
    assert hdf.get_keys(path) == []


def test_write_df(hdf_file_path, mock_key, mocker):
    df_mock = mocker.patch("vivarium.framework.artifact.hdf._write_pandas_data")
    data = pd.DataFrame(np.random.random((10, 3)), columns=["a", "b", "c"], index=range(10))

    hdf.write(hdf_file_path, mock_key, data)

    df_mock.assert_called_once_with(hdf_file_path, mock_key, data)


def test_write_json(hdf_file_path, mock_key, json_data, mocker):
    json_mock = mocker.patch("vivarium.framework.artifact.hdf._write_json_blob")
    hdf.write(hdf_file_path, mock_key, json_data)
    json_mock.assert_called_once_with(hdf_file_path, mock_key, json_data)


def test_load(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    data = hdf.load(hdf_file_path, key, filter_terms=None, column_filters=None)
    if "restrictions" in key or "versions" in key:
        assert isinstance(data, dict)
    elif "metadata" in key:
        assert isinstance(data, list)
    else:
        assert isinstance(data, pd.DataFrame)


def test_load_with_invalid_filters(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    data = hdf.load(hdf_file_path, key, filter_terms=["fake_filter==0"], column_filters=None)
    if "restrictions" in key or "versions" in key:
        assert isinstance(data, dict)
    elif "metadata" in key:
        assert isinstance(data, list)
    else:
        assert isinstance(data, pd.DataFrame)


def test_load_with_valid_filters(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    data = hdf.load(hdf_file_path, key, filter_terms=["year == 2006"], column_filters=None)
    if "restrictions" in key or "versions" in key:
        assert isinstance(data, dict)
    elif "metadata" in key:
        assert isinstance(data, list)
    else:
        assert isinstance(data, pd.DataFrame)
        if "year" in data.columns:
            assert set(data.year) == {2006}


def test_load_filter_empty_data_frame_index(hdf_file_path):
    key = hdf.EntityKey("cause.test.prevalence")
    data = pd.DataFrame(data={"age": range(10), "year": range(10), "draw": range(10)})
    data = data.set_index(list(data.columns))

    hdf._write_pandas_data(hdf_file_path, key, data)
    loaded_data = hdf.load(
        hdf_file_path, key, filter_terms=["year == 4"], column_filters=None
    )
    loaded_data = loaded_data.reset_index()
    assert loaded_data.year.unique() == 4


def test_remove(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    hdf.remove(hdf_file_path, key)
    with tables.open_file(str(hdf_file_path)) as file:
        assert key.path not in file


def test_get_keys(hdf_file_path, hdf_keys):
    assert sorted(hdf.get_keys(hdf_file_path)) == sorted(hdf_keys)


def test_write_json_blob(hdf_file_path, mock_key, json_data):
    hdf._write_json_blob(hdf_file_path, mock_key, json_data)

    with tables.open_file(str(hdf_file_path)) as file:
        node = file.get_node(mock_key.path)
        with filenode.open_node(node) as file_node:
            data = json.load(file_node)
            assert data == json_data


def test_write_empty_data_frame(hdf_file_path):
    key = hdf.EntityKey("cause.test.prevalence")
    data = pd.DataFrame(columns=("age", "year", "sex", "draw", "location", "value"))

    with pytest.raises(ValueError):
        hdf._write_pandas_data(hdf_file_path, key, data)


def test_write_empty_data_frame_index(hdf_file_path):
    key = hdf.EntityKey("cause.test.prevalence")
    data = pd.DataFrame(data={"age": range(10), "year": range(10), "draw": range(10)})
    data = data.set_index(list(data.columns))

    hdf._write_pandas_data(hdf_file_path, key, data)
    written_data = pd.read_hdf(hdf_file_path, key.path)
    written_data = written_data.set_index(
        list(written_data)
    )  # write resets index. only calling load undoes it
    assert written_data.equals(data)


def test_write_load_empty_data_frame_index(hdf_file_path):
    key = hdf.EntityKey("cause.test.prevalence")
    data = pd.DataFrame(data={"age": range(10), "year": range(10), "draw": range(10)})
    data = data.set_index(list(data.columns))

    hdf._write_pandas_data(hdf_file_path, key, data)
    loaded_data = hdf.load(hdf_file_path, key, filter_terms=None, column_filters=None)
    assert loaded_data.equals(data)


def test_write_data_frame(hdf_file_path):
    key = hdf.EntityKey("cause.test.prevalence")
    data = build_table(
        lambda x: random.choice([0, 1]),
        key_columns={"draw": [0, 1], "location": ["Kenya"]},
    )

    non_val_columns = data.columns.difference({"value"})
    data = data.set_index(list(non_val_columns))

    hdf._write_pandas_data(hdf_file_path, key, data)

    written_data = pd.read_hdf(hdf_file_path, key.path)
    assert written_data.equals(data)

    filter_terms = ["draw == 0"]
    written_data = pd.read_hdf(hdf_file_path, key.path, where=filter_terms)
    assert written_data.equals(data.xs(0, level="draw", drop_level=False))


def test_get_keys_private(hdf_file, hdf_keys):
    assert sorted(hdf._get_keys(hdf_file.root)) == sorted(hdf_keys)


def test_get_node_name(hdf_file, hdf_key):
    key = hdf.EntityKey(hdf_key)
    assert hdf._get_node_name(hdf_file.get_node(key.path)) == key.measure


def test_get_valid_filter_terms_all_invalid(hdf_key, hdf_file):
    node = hdf_file.get_node(hdf.EntityKey(hdf_key).path)
    if not isinstance(node, tables.earray.EArray):
        columns = node.table.colnames
        invalid_filter_terms = _construct_no_valid_filters(columns)
        assert hdf._get_valid_filter_terms(invalid_filter_terms, columns) is None


def test_get_valid_filter_terms_all_valid(hdf_key, hdf_file):
    node = hdf_file.get_node(hdf.EntityKey(hdf_key).path)
    if not isinstance(node, tables.earray.EArray):
        columns = node.table.colnames
        valid_filter_terms = _construct_all_valid_filters(columns)
        assert set(hdf._get_valid_filter_terms(valid_filter_terms, columns)) == set(
            valid_filter_terms
        )


def test_get_valid_filter_terms_some_valid(hdf_key, hdf_file):
    node = hdf_file.get_node(hdf.EntityKey(hdf_key).path)
    if not isinstance(node, tables.earray.EArray):
        columns = node.table.colnames
        invalid_filter_terms = _construct_no_valid_filters(columns)
        valid_filter_terms = _construct_all_valid_filters(columns)
        all_terms = invalid_filter_terms + valid_filter_terms
        result = hdf._get_valid_filter_terms(all_terms, columns)
        assert set(result) == set(valid_filter_terms)


def test_get_valid_filter_terms_no_terms():
    assert hdf._get_valid_filter_terms(None, []) is None


def _construct_no_valid_filters(columns):
    fake_cols = [
        c[1:] for c in columns
    ]  # strip out the first char to make a list of all fake cols
    terms = [c + " <= 0" for c in fake_cols]
    return _complicate_terms_to_parse(terms)


def _construct_all_valid_filters(columns):
    terms = [
        c + "=0" for c in columns
    ]  # assume c is numeric - we won't actually apply filter
    return _complicate_terms_to_parse(terms)


def _complicate_terms_to_parse(terms):
    n_terms = len(terms)
    if n_terms > 1:
        # throw in some parens and ifs/ands
        term_1 = "(" + " & ".join(terms[: (n_terms // 2 + n_terms % 2)]) + ")"
        term_2 = "(" + " | ".join(terms[(n_terms // 2 + n_terms % 2) :]) + ")"
        terms = [term_1, term_2] + terms
    return ["(" + t + ")" for t in terms]


def test_EntityKey_init_failure():
    bad_keys = ["hello", "a.b.c.d", "", ".", ".coconut", "a.", "a..c"]

    for k in bad_keys:
        error_msg = f'Invalid format for HDF key: {k}. Acceptable formats are "type.name.measure" and "type.measure"'
        with pytest.raises(ValueError, match=error_msg):
            hdf.EntityKey(k)


def test_EntityKey_no_name():
    type_ = "population"
    measure = "structure"
    key = hdf.EntityKey(f"{type_}.{measure}")

    assert key.type == type_
    assert key.name == ""
    assert key.measure == measure
    assert key.group_prefix == "/"
    assert key.group_name == type_
    assert key.group == f"/{type_}"
    assert key.path == f"/{type_}/{measure}"
    assert key.with_measure("age_groups") == hdf.EntityKey("population.age_groups")


def test_EntityKey_with_name():
    type_ = "cause"
    name = "diarrheal_diseases"
    measure = "incidence"
    key = hdf.EntityKey(f"{type_}.{name}.{measure}")

    assert key.type == type_
    assert key.name == name
    assert key.measure == measure
    assert key.group_prefix == f"/{type_}"
    assert key.group_name == name
    assert key.group == f"/{type_}/{name}"
    assert key.path == f"/{type_}/{name}/{measure}"
    assert key.with_measure("prevalence") == hdf.EntityKey(f"{type_}.{name}.prevalence")


def test_entity_key_equality():
    type_ = "cause"
    name = "diarrheal_diseases"
    measure = "incidence"
    string = f"{type_}.{name}.{measure}"
    key = hdf.EntityKey(string)

    class NonString:
        def __str__(self):
            return string

    nonstring = NonString()

    assert (
        key == string
    ), "Comparision using __eq__ between string object and equivalent EntityKey failed"
    assert not (
        key != string
    ), "Comparision using __ne__ between string object and equivalent EntityKey failed"
    assert (
        key != nonstring
    ), "Comparision using __eq__ between non-string object and equivalent EntityKey failed"
    assert not (
        key == nonstring
    ), "Comparision using __ne__ between non-string object and equivalent EntityKey failed"

    measure = "prevalence"
    string = f"{type_}.{name}.{measure}"

    assert (
        key != string
    ), "Comparision using __eq__ between string object and different EntityKey failed"
    assert not (
        key == string
    ), "Comparision using __ne__ between string object and different EntityKey failed"
    assert (
        key != nonstring
    ), "Comparision using __eq__ between non-string object and different EntityKey failed"
    assert not (
        key == nonstring
    ), "Comparision using __ne__ between non-string object and different EntityKey failed"
